# -*- coding: utf-8 -*-

import logging
from hashlib import md5


class Policy(object):

    kind = NotImplemented

    def __init__(self, factory, next_policy=None, service='bpolicy', logger=None):
        self.factory = factory
        self.next_policy = next_policy
        self.service = service
        logger_name = '{service}.{kind}'.format(service=service, kind=self.kind)
        self.logger = logger if logger else logging.getLogger(logger_name)
        self.logger_handler = logging.NullHandler()
        self.logger.addHandler(self.logger_handler)

    def discount_quota(self, discount):
        if self.next_policy:
            self.logger.debug('discount next_policy quota')
            self.next_policy.discount_quota(discount)

    def _gen_factory_signature(self):
        return md5(str(self.factory.__dict__)).hexdigest()[:7]

    def _gen_store_key(self, store_prefix, identity):
        return '{store_prefix}{service}:{kind}:{factory_signature}:{identity}'.format(
            store_prefix=self.factory.store_prefix,
            service=self.service,
            kind=self.kind,
            factory_signature=self._gen_factory_signature(),
            identity=identity)

    def check_policy(self, identity):
        if self.next_policy:
            self.logger.debug('check next policy')
            self.next_policy.check_policy(identity)


class PolicyFactory(object):

    policy_class = NotImplemented

    def instance(self, next_policy=None, **kwargs):
        return self.policy_class(self, next_policy, **kwargs)



# -*- coding: utf-8 -*-


__all__ = ['PolicyError',
           'Policy', 'PolicyFactory',
           'RatedPolicy', 'RatedPolicyFactory',
           'TimedPolicy', 'TimedPolicyFactory',
           'CERNetPolicy', 'CERNetPolicyFactory',
           'GenerationedPolicy', 'GenerationedPolicyFactory',
           'SECOND', 'MINUTE', 'HOUR', 'DAY', 'WEEK', 'POLICY_KIND', 'DEFAULT_LOGGER_FORMATER',
           'chained_policy', 'silent_check', 'is_private_ipaddr', 'is_cernet_ipaddr', 'load_cernet_data', 'load_cernet_data']


from .error import PolicyError                         # NOQA
from .store import Store, FakeStore                    # NOQA
from .base import Policy, PolicyFactory                # NOQA
from .rated import RatedPolicy, RatedPolicyFactory     # NOQA
from .timed import TimedPolicy, TimedPolicyFactory     # NOQA
from .cernet import CERNetPolicy, CERNetPolicyFactory  # NOQA
from .generationed import GenerationedPolicy, GenerationedPolicyFactory   # NOQA
from .consts import SECOND, MINUTE, HOUR, DAY, WEEK, POLICY_KIND, DEFAULT_LOGGER_FORMATER  # NOQA
from .utils import chained_policy, silent_check, is_private_ipaddr, is_cernet_ipaddr, load_cernet_data  # NOQA



# -*- coding: utf-8 -*-

import csv
import IPy
import os.path
from .error import PolicyError


def chained_policy(*factories):
    policy = None
    for factory in factories:
        policy = factory.instance(policy)
    return policy


def silent_check(policy, identity):
    success = True
    try:
        policy.check_policy(identity)
    except PolicyError:
        success = False
    return success


class memoize(object):

    def __init__ (self, func):
        self.func = func
        self.mem = {}

    def __call__ (self, *args, **kwargs):
        if (args, str(kwargs)) in self.mem:
            result = self.mem[args, str(kwargs)]
        else:
            result = self.func(*args, **kwargs)
            self.mem[args, str(kwargs)] = result
        return result


@memoize
def load_cernet_data():
    cernet_resource = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/cernet_ip.csv')
    cernet_nets = [r[0] for r in csv.reader(open(cernet_resource), delimiter='\t')]
    return IPy.IPSet([IPy.IP('%s/%s' % (ip_net, (ip_net.count('.') + 1) * 8), make_net=True) for ip_net in cernet_nets])


def is_cernet_ipaddr(ipstr):
    try:
        return IPy.IP(ipstr) in load_cernet_data()
    except ValueError:
        pass


def is_private_ipaddr(ipstr):
    try:
        return IPy.IP(ipstr).iptype() == 'PRIVATE'
    except ValueError:
        pass



# -*- coding: utf-8 -*-

from datetime import datetime

from .consts import POLICY_KIND
from .base import Policy, PolicyFactory


class TimedPolicy(Policy):

    kind = POLICY_KIND.TIMED

    def get_current_time(self):
        return datetime.now().time()

    def check_policy(self, identity):
        if self.factory.start_time < self.get_current_time() < self.factory.end_time:
            self.logger.debug('timed policy(%s to %s) encountered, discount all successor policy quota by %s', self.factory.start_time, self.factory.end_time, self.factory.discount)
            self.discount_quota(self.factory.discount)
        else:
            self.logger.debug('timed policy not encountered, silent pass through')
        super(TimedPolicy, self).check_policy(identity)


class TimedPolicyFactory(PolicyFactory):

    policy_class = TimedPolicy

    def __init__(self, start_time, end_time, discount):
        self.start_time = start_time
        self.end_time = end_time
        self.discount = discount



# -*- coding: utf-8 -*-

from .consts import POLICY_KIND
from .error import PolicyError
from .base import Policy, PolicyFactory


class RatedPolicy(Policy):

    kind = POLICY_KIND.RATED

    def __init__(self, factory, next_policy=None):
        self.quota = factory.quota
        super(RatedPolicy, self).__init__(factory, next_policy)

    def discount_quota(self, discount):
        origin_quota = self.quota
        self.quota = int(self.quota * discount)
        self.logger.debug('discount current quota from %s to %s', origin_quota, self.quota)
        super(RatedPolicy, self).discount_quota(discount)

    def check_policy(self, identity):
        store_key = self._gen_store_key(self.factory.store_prefix, identity)
        current_counter = self.factory.store.get(store_key)
        if current_counter is None:
            current_counter = 1
            self.factory.store.set(store_key, 1, self.factory.interval)
            self.logger.debug('initial current_counter to 1 for new interval')
        else:
            current_counter += 1
            self.factory.store.incr(store_key, 1)
            self.logger.debug('incr current_counter from %s to %s', current_counter, current_counter + 1)
        if current_counter > self.quota:
            self.logger.debug('max quota encountered: %s / %s', current_counter, self.quota)
            raise PolicyError(self, 'max quota exceed')
        super(RatedPolicy, self).check_policy(identity)


class RatedPolicyFactory(PolicyFactory):

    policy_class = RatedPolicy

    def __init__(self, interval, quota, store, store_prefix=''):
        self.interval = interval
        self.quota = quota
        self.store = store
        self.store_prefix = store_prefix



# -*- coding: utf-8 -*-

from time import time


class Store(object):

    def get(self, key):
        raise NotImplementedError

    def set(self, key, value, expire=None):
        raise NotImplementedError

    def incr(self, key, step=1):
        raise NotImplementedError


class FakeStore(Store):

    def __init__(self, *args, **kwargs):
        self.cache = {}
        self.current_timestamp = int(time())

    def get(self, key):
        self._check_expire()
        return self.cache.get(key, (0, None))[1]

    def set(self, key, value, expire=None, min_compress_len=0):
        timeout = self._get_current_timestamp() + expire if expire else None
        self.cache[key] = (timeout, value)
        return True

    def incr(self, key, step=1):
        value = self.get(key)
        if value is None:
            return None
        new_value = value + step
        self.cache[key] = (self.cache[key][0], new_value)
        return new_value

    def delete(self, key):
        if key in self.cache:
            del self.cache[key]

    def _clear_cache(self, prefix=''):
        for key in self.cache.keys():
            if key.startswith(prefix):
                del self.cache[key]

    def _check_expire(self):
        for k in self.cache.keys():
            timeout, value = self.cache[k]
            if timeout and self._get_current_timestamp() >= timeout:
                del self.cache[k]

    def _get_current_timestamp(self):
        return self.current_timestamp

    def _set_current_timestamp(self, timestamp):
        self.current_timestamp = timestamp

    def _incr_current_timestamp(self, step):
        self.current_timestamp += step



# -*- coding: utf-8 -*-


class PolicyError(Exception):

    def __init__(self, policy, message):
        self.policy = policy
        super(PolicyError, self).__init__(message)

    def __str__(self):
        return '{class_name}({policy}: {message})'.format(class_name=self.__class__.__name__, policy=self.policy.kind, message=self.message)



# -*- coding: utf-8 -*-

from .consts import POLICY_KIND
from .base import Policy, PolicyFactory
from .utils import load_cernet_data, is_cernet_ipaddr


class CERNetPolicy(Policy):

    kind = POLICY_KIND.CERNET

    def __init__(self, factory, next_policy=None):
        self.discount = factory.discount
        super(CERNetPolicy, self).__init__(factory, next_policy)

    def check_policy(self, identity):
        if is_cernet_ipaddr(identity):
            self.logger.debug('%r is valid cernet ipaddress, silent pass through', identity)
        else:
            self.logger.debug('%r is not valid cernet ipaddress, discount all successor policy quota by %s', identity, self.discount)
            self.discount_quota(self.discount)
        super(CERNetPolicy, self).check_policy(identity)


class CERNetPolicyFactory(PolicyFactory):

    policy_class = CERNetPolicy

    def __init__(self, discount):
        self.discount = discount
        load_cernet_data()



# -*- coding: utf-8 -*-

import time

from .consts import POLICY_KIND
from .error import PolicyError
from .base import Policy, PolicyFactory


class GenerationedPolicy(Policy):

    kind = POLICY_KIND.GENERATIONED

    def __init__(self, factory, next_policy=None):
        self.quota = factory.quota
        super(GenerationedPolicy, self).__init__(factory, next_policy)

    def discount_quota(self, discount):
        origin_quota = self.quota
        self.quota = int(self.quota * discount)
        self.logger.debug('discount current quota from %s to %s', origin_quota, self.quota)
        super(GenerationedPolicy, self).discount_quota(discount)

    def get_current_timestamp(self):
        return int(time.time())

    def check_policy(self, identity):
        stats_store_key = self._gen_store_key(self.factory.store_prefix, identity)
        current_discount, current_counter, current_timestamp = (1, 1, self.get_current_timestamp())
        latest_discount, latest_counter, latest_timestamp = self.factory.store.get(stats_store_key) or (current_discount, current_counter - 1, current_timestamp)
        latest_period, current_period = latest_timestamp // self.factory.interval, current_timestamp // self.factory.interval

        if latest_period == current_period:
            current_counter = latest_counter + 1
            current_discount = latest_discount
            self.logger.debug('incr current_counter from %s to %s, keep current_discount %s', current_counter - 1, current_counter, current_discount)
        elif current_period - latest_period < self.factory.max_keep_traking:
            current_discount = latest_discount * self.factory.discount if latest_counter > self.quota * latest_discount else latest_discount / self.factory.discount
            current_discount = min(current_discount, 1)
            self.logger.debug('max_keep_traking encountered, current_discount to %s', current_discount)
        else:
            self.logger.debug('initial a new generation')

        self.factory.store.set(stats_store_key, (current_discount, current_counter, current_timestamp), self.factory.interval * self.factory.max_keep_traking)

        if current_counter > self.quota * current_discount:
            self.logger.debug('max quota encountered: %s / %s', current_counter, self.quota * current_discount)
            raise PolicyError(self, 'max quota exceed')

        super(GenerationedPolicy, self).check_policy(identity)


class GenerationedPolicyFactory(PolicyFactory):

    policy_class = GenerationedPolicy

    def __init__(self, quota, interval, discount, max_keep_traking, store, store_prefix=''):
        self.quota = quota
        self.interval = interval
        self.discount = discount
        self.max_keep_traking = max_keep_traking
        self.store = store
        self.store_prefix = store_prefix



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='bpolicy',
    version='0.1.1',
    author='mapix',
    author_email='mapix.me@gmail.com',
    packages=find_packages(),
    url='https://github.com/mapix/bpolicy',
    description='A collection of basic policies for web application anti-spammer use.',
    long_description=open('README.rst').read(),
    install_requires=['IPy==0.81'],
    license='MIT'
)



# -*- coding: utf-8 -*-

import logging

from datetime import timedelta
from collections import namedtuple

SECOND = int(timedelta(seconds=1).total_seconds())
MINUTE = int(timedelta(minutes=1).total_seconds())
HOUR = int(timedelta(hours=1).total_seconds())
DAY = int(timedelta(days=1).total_seconds())
WEEK = int(timedelta(weeks=1).total_seconds())

POLICY_KIND = namedtuple('POLICY_KIND', ['TIMED', 'RATED', 'CERNET', 'GENERATIONED'])('timed', 'rated', 'cernet', 'generationed')
DEFAULT_LOGGER_FORMATER = logging.Formatter(fmt='[%(asctime)s]\t[%(levelname)s]\t%(name)-20s\t%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
